#!/usr/bin/env python
#
# Program:     $Id: tbt.py 442 2012-11-08 08:02:01Z wdbrinkm $
# Description: Too-Big-Trick library
# 
# Too-Big_Trick library provides an IP6 fingerprinting technique to assist in
# router-level alias resolution by comparing fragment identification numbers
# received from targeted router interfaces.  An ICMP6 Too Big message is sent 
# invoking source router fragmentation iliciting a unique identifier common to 
# all interfaces of a router.  Received fragmentation identification numbers 
# are compared to determine if provided IP6 addresses are aliases.
#
# Copyright (c) 2012 William Brinkmeyer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import getopt
from time import sleep
from scapy.all import *

class TBT:
    # Initialize class variables.
    def __init__(self, src_addr, tunnel_addr=None, _verbose=False):
        self.verbose = _verbose
        self.src_addr = src_addr
        self.dst_addr = 0
        self.tunnel_addr = tunnel_addr
        self.echo_seq = 0
        self.frag_id = []
        self.fullmtu = 1300 # Set to be large, but small enough for a tunnel 

    # Assigns initial IP6 destination.  
    def destination(self, _dst_addr):
        self.dst_addr = _dst_addr

    # Sends ICMP6 echo request to specified address.  Expected responses vary 
    # based on execution in relation to ICMP6 Too Big message.  Expected 
    # response is non-fragmented response prior to sending ICMP6 Too Big 
    # message.  Following ICMP6 Too Big message, expected response is 
    # fragmented ICMP6 echo reply from target router. 
    def echo_req(self, length, dst_addr=None):
        if (dst_addr == None):
            dst_addr = self.dst_addr
        data_string = RandString(length)
        self.echo_seq += 1
        if (self.tunnel_addr):
            send(IP(dst = self.tunnel_addr, proto = 41)/\
                IPv6(src = self.src_addr, dst = dst_addr, hlim = 55, nh = 58)/\
                ICMPv6EchoRequest(type = 128, seq = self.echo_seq, code = 0, \
                    data = data_string), verbose = self.verbose)
        else:
            send(IPv6(src = self.src_addr, dst = dst_addr, hlim = 55, \
                nh = 58) / ICMPv6EchoRequest(type = 128, seq = self.echo_seq, \
                code = 0, data = data_string), verbose = self.verbose)

    # Utilizes scapy function to send ICMP6 Too Big message to IP6 destination
    # address.  Expected result of function is for sent packet to force target
    # to update Path MTU for route enabling returned traffic larger than 1280
    # bytes to be fragmented.
    def too_big(self, dst_addr):
        data_string = RandString(1280-96)
        if (self.tunnel_addr): 
            send(IP(dst = self.tunnel_addr, proto = 41)/\
  	        IPv6(src = self.src_addr, dst = dst_addr, hlim = 55, nh = 58)/\
  	        ICMPv6PacketTooBig(mtu = 1280, type = 2, code = 0)/\
  	        (IPv6(src = dst_addr, dst = self.src_addr, hlim = 1)/\
  	        ICMPv6EchoReply(type = 129, seq = self.echo_seq, code = 0, \
                data = data_string)), verbose = self.verbose)
        else:
            send(IPv6(src = self.src_addr, dst = dst_addr, hlim = 55, \
                nh = 58) / ICMPv6PacketTooBig(mtu = 1280, type = 2, code = 0)/\
                (IPv6(src = dst_addr, dst = self.src_addr, hlim = 1)/\
                ICMPv6EchoReply(type = 129, seq = self.echo_seq, code = 0, \
                data = data_string)), verbose = self.verbose)

    # Compares fragment identification numbers from two IP6 addresses stored in
    # self.frag_id to determine if fragment identification numbers are 
    # sequential.  If all values in self.frag_id are sequential, probed IP6
    # addresses are identified as aliases.  
    def is_alias(self):
        i = 0
        while ((i + 1) < len(self.frag_id)):
            if (self.frag_id[i]+1 == self.frag_id[i+1]):
                i += 1
            else:
                return False
        return True
 
    # Store fragment identification number received from probed IP6 destination 
    def store_frag(self, pkt):
        if (pkt[IPv6].nh == 44):
            self.frag_id.append(pkt[IPv6ExtHdrFragment].id)

    # Single host verification for TBT.  Outputs a list of received fragment 
    # identification numbers.
    def verify_TBT(self):
        self.echo_req(self.fullmtu)
        sleep(0.1)
        self.too_big(self.dst_addr)
        sleep(0.2)
        if (self.tunnel_addr):
            for i in range(5):
                self.echo_req(self.fullmtu)
                sniff(filter = "ip proto 41", prn = self.store_frag, \
                    count = 1, timeout = 2)
                sleep(0.1)
        else:
            for i in range(5):
                self.echo_req(self.fullmtu)
                sniff(filter = "ip6 proto 44", prn = self.store_frag, \
                    count = 1, timeout = 2)
                sleep(0.1)
        print ("Fragment IDs for %s are: %s" %(self.dst_addr,self.frag_id))

    # Sends initial ICMP6 Echo Request, then performs Too Big Trick, then sends
    # ICMP6 echo request large enough to cause destination to fragment 
    # responses (ICMP6 Too Big message sent prior to calling resolve()).  Alias
    # comparison checks are performed throughout function to minimize the 
    # amount of generated traffic.
    def resolve(self, alias):
        self.echo_req(self.fullmtu)
        sleep(0.1)
        self.too_big(self.dst_addr)
        self.echo_req(self.fullmtu, alias)
        sleep(0.1)
        self.too_big(alias)
        sleep(0.2)
        if (self.tunnel_addr):
            for i in range(5):
                self.echo_req(self.fullmtu)
                sniff(filter = "ip proto 41", prn = self.store_frag, \
                    count = 1, timeout = 2)
                sleep(0.1)
                self.echo_req(self.fullmtu, alias)
                sniff(filter = "ip proto 41", prn = self.store_frag, \
                    count = 1, timeout = 2)
                sleep(0.1)
                if (not self.is_alias()):
                    return False
                self.echo_req(self.fullmtu)
                sniff(filter = "ip proto 41", prn = self.store_frag, \
                    count = 1, timeout = 2)
                sleep(0.1)
                if (not self.is_alias()):
                    return False
        else:
            for i in range(5):
                self.echo_req(self.fullmtu)
                sniff(filter = "ip6 proto 44", prn = self.store_frag, \
                    count = 1, timeout = 2)
                sleep(0.1)
                self.echo_req(self.fullmtu, alias)
                sniff(filter = "ip6 proto 44", prn = self.store_frag, \
                    count = 1, timeout = 2)
                sleep(0.1)
                if (not self.is_alias()):
                    return False
                self.echo_req(self.fullmtu)
                sniff(filter = "ip6 proto 44", prn = self.store_frag, \
                    count = 1, timeout = 2)
                sleep(0.1)
                if (not self.is_alias()):
                    return False
        return True


# Display valid switches to user
def help(args):
    print args[0], ": [--source][--destination][--alias][--tunnel]"
    sys.exit(0)


if __name__ == "__main__":
    try:
        opts, args = getopt.getopt(sys.argv[1:], "s:d:a:t:h", \
            ["source=", "destination=", "alias=", "tunnel="])
    except getopt.GetoptError, err:
        print str(err)
        help(sys.argv)

    source = None
    tunnel = None
    dest   = None
    alias  = None
    for o, a in opts:
        if o in ("-s", "--source"):
            source = a
        elif o in ("-d", "--destination"):
            dest = a
        elif o in ("-a", "--alias"):
            alias = a
        elif o in ("-t", "--tunnel"):
            tunnel = a
        elif o in ("-h", "--help"):
            help(sys.argv)
        else:
            assert False, "unhandled option"
    if (not source) or (not dest):
        help(sys.argv)

    # Send ICMP6 Too Big message via tunnel end point, if declared, or
    # directly to IP6 destination.
    if (tunnel):
        t = TBT(source, tunnel)
    else:
        t = TBT(source)
    t.destination(dest)
    t.echo_req(1300)
    # If two IP6 addresses provided, perform alias resolution, else verify
    # target response to Too Big Trick.
    if (alias):
        res = t.resolve(alias)
        if (res):
            print dest, "and", alias, "are aliases."
        else:
            print dest, "and", alias, "NOT aliases."
    else:
        t.verify_TBT()
